
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>
#include <errno.h>
#include <assert.h>

#include <iostream>
#include <string>
#include <vector>
#include <cstdio>
#include <fstream>

using namespace std;

typedef unsigned short bu16;
typedef unsigned long bu32;

// this is what Posix says, not counting null
#define kMaxPathLen 255

typedef struct file_attrs {
	char path[kMaxPathLen+1];
	bu16 mode, uid, gid;
	long long size;
	bu32 ctime, mtime;//, atime;
	bu32 pos; // offset of file data in archive file if regular file
//	int incFlag; // this flag is used during an incremental backup
		// to indicate that this file should be written into
		// the new archive.
	struct file_attrs *next; // for the temp list
} file_attrs;

// this is our error reporting function.
void report_err(int inErr, const char *inDesc);
// this fills in our file_attrs record for
// a file with the given pathname
int get_file_attrs(char *inPath, file_attrs *outAttrs);

int write_manifest(const vector<string>& manifest, const char *manifestBase, int mi);
int build_manifests(long long mediaSize, const char *manifestBase);
// write out an entire file
int writeFile(const char *inPathStr, void *inPtr, long inSize);

////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	if( argc < 3 )
	{
		cerr << "Usage: flashback <mediasize> <manifestbasename>" << endl;
		return 2;
	}

	long long mediaSize = atoll(argv[1]);
	const char *manifestBase = argv[2];

	int err = build_manifests(mediaSize,manifestBase);

	return err == 0;
}

int write_manifest(const vector<string>& manifest, const char *manifestBase, int mi)
{
	ofstream out;
	try
	{
		string path(manifestBase);
		char nstr[100];
		sprintf(nstr,"%d",mi);
		path += nstr;

		out.open(  path.c_str(),  ios::out );

		// out << "Manifest " << mi << endl;
		vector<string>::const_iterator it;
		for(it=manifest.begin();it!=manifest.end();it++)
			out << *it << endl;
		// out << "-------" << endl;

		out.close();
	}
	catch(...)
	{
		report_err(-1,"write_manifest");
		return -1;
	}
	
	return 0;
}

int build_manifests(long long mediaSize, const char *manifestBase)
{
	// First task is to create manifests.
	char path[kMaxPathLen+2];
	int slen;
	int err = 0;
	file_attrs attrs;

	int mi = 0;
	long long total = 0;
	vector<string> manifest;

	while( fgets( path, kMaxPathLen+1, stdin ) )
	{
		slen = strlen(path);
		// kill the newline on the end
		if( path[slen-1] == '\n' )
			path[slen-1] = 0; 
		err = get_file_attrs(path,&attrs);
		if( err )
		{
			report_err(err,path);
			// keep going.
		}
		else
		{
			printf("path = %s, size = %lld\n",path,(long long)attrs.size);
			// NOTE: add to manifest if will fit,
			// if not write out manifest and start new one.

			// always add atleast one file.
			long long remaining = mediaSize - total;
			if( remaining < attrs.size )
			{
				cout << "total: " << total << endl;
				err = write_manifest(manifest,manifestBase,++mi);
				if( err )
				{
					report_err(err,manifestBase);
					// break; // fatal if can't write manifest?
				}
				manifest.clear();
				total = 0;
			}
			manifest.push_back( string( path ) );
			total += attrs.size;
			
		}

	}

	// write the leftovers
	if( manifest.size() != 0 )
	{
		cout << "total: " << total << endl;
		err = write_manifest(manifest, manifestBase,++mi);
		if( err )
		{
			report_err(err,manifestBase);
			// break; // fatal if can't write manifest?
		}
	}

	// NOTE: make exit status nonzero if had errors?
	return 0;
}

// this is our error reporting function.
void report_err(int inErr, const char *inDesc)
{
	fprintf(stderr,"Error: %s %s(%d)\n",inDesc,strerror(inErr),inErr);
}

// this fills in our file_attrs record for
// a file with the given pathname
int get_file_attrs(char *inPath, file_attrs *outAttrs)
{
	struct stat st_buf;

	assert( inPath );
	assert( outAttrs );
	assert( strlen(inPath) <= kMaxPathLen );

	if( stat( inPath, &st_buf ) )
		return errno;

	// zero out the path field.
	memset( outAttrs->path, 0, kMaxPathLen+1 );
	strncpy( outAttrs->path, inPath, kMaxPathLen );
//	outAttrs->path[kMaxPathLen] = 0; // won't be null terminated
		// otherwise if strncpy hit max
	outAttrs->mode = st_buf.st_mode;
	outAttrs->uid = st_buf.st_uid;
	outAttrs->gid = st_buf.st_gid;
	outAttrs->size = st_buf.st_size;
	outAttrs->ctime = st_buf.st_ctime;
	outAttrs->mtime = st_buf.st_mtime;
//	outAttrs->atime = st_buf.st_atime;
	outAttrs->pos = (bu32)-1L; // flag it as not set yet

	return 0;
}


